////////////////////////////////////////////////////////////////
// If this code works, it was written by Alexander Nazarenko.
// If not, I don't know who wrote it.

//   ⨢ True-Cpp plugin. (c) Andrey Tretjakov,1999 v.1.0

#include "winmem.h"

#define _FAR_USE_FARFINDDATA

#include <plugin.hpp>
#include <farkeys.hpp>

#include "eicoll.h"
#include "true-tpl.h"
#include "ttpl-lng.h"
#include "farintf.h"
#include "console.h"
#include "lstring.h"

#include "cregexp.h"

#include "syslog.h"

bool IsOldFar = true;

#define expAnyWhere ".*\\b\\p.*"
#define expAtStart  "\\p.*"
#define expAtMiddle "\\S.*\\b\\p.*\\S"
#define expAtEnd    ".*\\b\\p"
#define expOnBlank  "\\p"

#define MAX_REG_LEN 128
#define MAX_STR_LEN 4096

static int defExpandFKey = KEY_SPACE;

struct TMacro
{
  int FARKey;
  bool atStartup;
  char Word[MAX_REG_LEN], At[MAX_REG_LEN], MacroText[MAX_STR_LEN];
  char before[MAX_REG_LEN], after[MAX_REG_LEN], immChar;
};

struct TIndent
{
  char mask[MAX_REG_LEN], relative[MAX_REG_LEN], immChar;
  int indent[2], start, BracketsMode, bracketLink;
};

struct TBracket
{
  char open[MAX_REG_LEN], close[MAX_REG_LEN];
};

struct TComment
{
  char mask[MAX_REG_LEN];
};

struct TDefine
{
  TDefine(const char*, const char*);
  char name[MAX_REG_LEN], value[MAX_STR_LEN];
};

TDefine::TDefine(const char *aName, const char *aValue)
{
  lstrcpy(name, aName);
  lstrcpy(value, aValue);
}

enum TJumpType { jtNone, jtFirst, jtMenu, jtSmart };
enum TSaveType { saNone, saAll,  saCurrent };
enum TCDType   { cdNone, cdFile, cdBase };

struct TCompiler
{
  TCompiler();
  TCompiler(const TCompiler&);
  char title[MAX_REG_LEN], err[MAX_REG_LEN];
  int line, col, fileMatch;
  bool searchCol;
  void defaults(void)
  {
    *title = *err = 0;
    line = col = fileMatch = -1;
    searchCol = false;
  }
};

TCompiler::TCompiler()
{
  defaults();
}

TCompiler::TCompiler(const TCompiler& e)
{
  lstrcpy(title, e.title);
  lstrcpy(err, e.err);
  line = e.line;
  col = e.col;
  searchCol = e.searchCol;
  fileMatch = e.fileMatch;
}

struct TExec
{
  TExec();
  TExec(const TExec&);
  char title[MAX_REG_LEN], cmd[MAX_STR_LEN], compiler[MAX_STR_LEN];
  char enable[NM], disable[NM];
  TJumpType jumpType;
  TSaveType saveType;
  TCDType cd;
  bool searchBase, echo;
  void defaults(void)
  {
    *title = *cmd = *compiler = *enable = *disable = 0;
    saveType = saCurrent;
    jumpType = jtSmart;
    cd = cdNone;
    searchBase = false;
    echo = true;
  }
};

TExec::TExec()
{
  defaults();
}

TExec::TExec(const TExec& e)
{
  lstrcpy(title, e.title);
  lstrcpy(cmd, e.cmd);
  lstrcpy(compiler, e.compiler);
  lstrcpy(enable, e.enable);
  lstrcpy(disable, e.disable);
  jumpType = e.jumpType;
  saveType = e.saveType;
  cd = e.cd;
  echo = e.echo;
  searchBase = e.searchBase;
}

struct TLang
{
  TLang();
  char mask[MAX_REG_LEN], desc[MAX_REG_LEN], imm[MAX_REG_LEN], immExp[MAX_REG_LEN];
  bool ignoreCase;
//  bool panel;
  TExec defExec;
  TCollection macroColl, indentColl, bracketColl, commentColl, execColl, compilerColl, defineColl;
};

TLang::TLang()
{
  defineColl.insert(new TDefine("lt",   "<"));
  defineColl.insert(new TDefine("gt",   ">"));
  defineColl.insert(new TDefine("quot", "\""));
  defineColl.insert(new TDefine("amp",  "&"));
  defineColl.insert(new TDefine("nbsp", " "));
}

static void delLang(void *it)
{
  TLang *lng = (TLang*)it;
  lng->macroColl.removeAll();
  lng->indentColl.removeAll();
  lng->bracketColl.removeAll();
  lng->commentColl.removeAll();
  lng->execColl.removeAll();
  lng->compilerColl.removeAll();
  lng->defineColl.removeAll();
}

class TLangCollection: public TCollection
{
  public:
  TLangCollection() : TCollection(5, 5, delLang) {};
};

static TLangCollection langColl;
static int pluginBusy = 0;
static int pluginStop = 0;
static int scrollStop = 0;

static char *addExt(char *path, char *defExt)
{
  int n = lstrlen(path)-1;
  for ( int i = n ; i >= 0 ; i-- )
  {
    if ( path[i] == '.' )
    {
      lstrcpy(path+i, defExt);
      break;
    }
    if ( ( path[i] == '\\' ) || ( path[i] == ':' ) || ( i == 0 ) )
      return lstrcat(path, defExt);
  }
  return path;
}

static inline int isCharSpace(char c)
{
  return lstrchr(" \t\r\n", c) != NULL;
}

static char *skipSpaces(char*& line)
{
  while ( *line && isCharSpace(*line) )
    line++;
  return line;
}

static char *getWord(char*& line, char *kwd)
{
  char *k = kwd;
  if ( *skipSpaces(line) == '\"')
  {
    while ( *++line )
    {
      if ( *line == '\\' && line[1] == '\"' )
        *k++ = *++line;
      else if ( *line == '\"' )
      {
        if ( *skipSpaces(++line) != '\"' )
          break;
      }
      else
        *k++ = *line;
    }
    *k = 0;
  }
  else
  {
    while ( *line && !isCharSpace(*line) )
      *k++ = *line++;
    *k = 0;
  }
  return kwd;
}

static char *parseItem(TCollection *dc, char*& line, char *kwd, char *value)
{
  if ( IsCharAlpha(*skipSpaces(line)) )
  {
    char *k = kwd;
    while ( *line && IsCharAlpha(*line) )
      *k++ = *line++;
    *k = *value = 0;
    if ( *skipSpaces(line) == '=' )
    {
      line++;
      getWord(line, value);
      if ( dc )
      {
        char *v, pattern[MAX_REG_LEN], rest[MAX_STR_LEN];
        for ( size_t i = 0 ; i < dc->getCount() ; i++ )
        {
          TDefine *d = (TDefine*)((*dc)[i]);
          lstrcat(lstrcat(lstrcpy(pattern, "&"), d->name), ";");
          while ( ( v = lstrstr(value, pattern) ) != NULL )
          {
            *v = 0;
            lstrcpy(rest, v+lstrlen(pattern));
            lstrcat(lstrcat(v, d->value), rest);
          }
        }
      }
    }
    return line;
  }
  return NULL;
}

static char *getItem(char*& line, char *kwd, bool& group)
{
  group = true;
  if ( *skipSpaces(line) == '<')
  {
    char *stLine = line;
    char *k = kwd;
    line++;
    if ( *line && ( *line == '/') )
      *k++ = *line++;
    while ( *line && IsCharAlpha(*line) )
      *k++ = *line++;
    *k = 0;
    k = --line;
    int inQuote = 0;
    while ( *++line )
    {
      if ( inQuote )
      {
        if ( *line == '\\' && line[1] == '\"' )
          line++;
        else if ( *line == '\"' )
          inQuote = 0;
      }
      else
      {
        if ( *line == '>' )
        {
          if ( line > stLine )
          {
            --line;
            if ( *line == '/' )
            {
              *line = 0;
              group = false;
            }
            line++;
          }
          *line = 0;
          line++;
          break;
        }
        else if ( *line == '\"' )
          inQuote = 1;
      }
    }
    return k;
  }
  return NULL;
}

static void squeeze(char *path)
{
  char *dest = path;
  char *src = path;
  while ( *src != 0 )
  {
    if ( *src != '.' )
      *dest++ = *src++;
    else
    {
      src++;
      if ( *src == '.' )
      {
        src += 2;
        dest--;
        while ( *--dest != '\\' )
          ;
        dest++;
      }
      else
      {
        src++;
        dest += 2;
      }
    }
  }
  *dest = 0;
}

enum enSplitType { enDRIVE = 1, enDIRECTORY = 2, enFILENAME = 4, enEXTENTION = 8 };

static int fnSplit(const char *path, char *drive, char *dir, char *file, char *ext)
{
  int flag = *drive = *dir = *file = *ext = 0;
  char ch, *p = lstrchr((char*)path, ':');
  if ( p )
  {
    ch = p[1];
    p[1] = 0;
    lstrcpy(drive, path);
    flag |= enDRIVE;
    p[1] = ch;
    p++;
  }
  else
    p = (char*)path;
  char *dp = lstrrchr(p, '\\');
  if ( dp )
  {
    ch = dp[1];
    dp[1] = 0;
    lstrcpy(dir, p);
    dp[1] = ch;
    p = dp+1;
    flag |= enDIRECTORY;
  }
  dp = lstrchr(p, '.');
  if ( dp )
  {
    lstrcpy(ext, dp);
    flag |= enEXTENTION;
    ch = *dp;
    *dp = 0;
    lstrcpy(file, p);
    *dp = ch;
  }
  else
    lstrcpy(file, p);
  if ( *file  ) flag |= enFILENAME;
  return flag;
}

static inline char *fnMerge(char *path, char *drive, char *dir, char *file, char *ext)
{
  FSF.AddEndSlash(lstrcat(lstrcpy(path, drive), dir));
  return lstrcat(lstrcat(path, file), ext);
}

static char *fExpand(char *rpath, char *defPath = NULL)
{
  static char path[NM], drive[NM], dir[NM], file[NM], ext[NM];
  static char defDrive[4], defDir[NM];
  if ( defPath )
    lstrcpy(path, defPath);
  else
    GetCurrentDirectory(sizeof(path), path);
  FSF.AddEndSlash(path);
  fnSplit(path, defDrive, defDir, file, ext);
  int flags = fnSplit(rpath, drive, dir, file, ext);
  if ( (flags & enDRIVE) == 0 )
    lstrcpy(drive, defDrive);
  CharUpperBuff(drive, 1);
  if ( ( flags & enDIRECTORY ) == 0 || ( ( *dir != '\\' ) && ( *dir != '/' ) ) )
  {
    static char curdir[NM];
    lstrcpy(curdir, dir);
    lstrcat(lstrcpy(dir, defDir), curdir);
  }
  squeeze(dir);
  fnMerge(path, drive, dir, file, ext);
  return lstrcpy(rpath, path);
}

static char *getFile(char *path, char *fn)
{
  char file[NM], *fileBuff = NULL;
  HANDLE s = CreateFile(fExpand(lstrcpy(file, fn), path), GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
  if ( s != INVALID_HANDLE_VALUE )
  {
    unsigned fileSize = GetFileSize(s, NULL);
    if ( fileSize )
    {
      if ( ( fileBuff = new char[fileSize+1] ) != NULL )
      {
        DWORD rb = 0;
        if ( ReadFile(s, fileBuff, fileSize, &rb, NULL) && rb )
          fileBuff[rb] = 0;
        else
        {
          delete [] fileBuff;
          fileBuff = NULL;
        }
      }
    }
  }
  CloseHandle(s);
  return fileBuff;
}

static void parseExec(TExec *tmpe, const char *name, const char *value)
{
  if ( !FSF.LStricmp(name, "Save") )
  {
    if      ( !FSF.LStricmp(value, "Current") ) tmpe->saveType = saCurrent;
    else if ( !FSF.LStricmp(value, "All") )     tmpe->saveType = saAll;
    else                                        tmpe->saveType = saNone;
  }
  else if ( !FSF.LStricmp(name, "CD") )
  {
    if      ( !FSF.LStricmp(value, "File") )    tmpe->cd = cdFile;
    else if ( !FSF.LStricmp(value, "Base") )    tmpe->cd = cdBase;
    else                                        tmpe->cd = cdNone;
  }
  else if ( !FSF.LStricmp(name, "Jump") )
  {
    if      ( !FSF.LStricmp(value, "Smart") )   tmpe->jumpType = jtSmart;
    else if ( !FSF.LStricmp(value, "Menu") )    tmpe->jumpType = jtMenu;
    else if ( !FSF.LStricmp(value, "First") )   tmpe->jumpType = jtFirst;
    else                                        tmpe->jumpType = jtNone;
  }
  else if ( !FSF.LStricmp(name, "Echo") )       tmpe->echo = FSF.atoi(value) ? true : false;
  else if ( !FSF.LStricmp(name, "Title") )      lstrcpy(tmpe->title, value);
  else if ( !FSF.LStricmp(name, "Enable") )
  {
    if ( value[0] == '*' && value[1] == '\\' )
    {
      lstrcpy(tmpe->enable, value);
      tmpe->searchBase = true;
    }
    else
      lstrcpy(tmpe->enable, value+2);
  }
  else if ( !FSF.LStricmp(name, "Disable") )    lstrcpy(tmpe->disable, value);
  else if ( !FSF.LStricmp(name, "Compiler") )   lstrcpy(tmpe->compiler, value);
}

static void parseCompiler(TCompiler *tmpc, const char *name, const char *value)
{
  if      ( !FSF.LStricmp(name, "Error") )      lstrcpy(tmpc->err, value);
  else if ( !FSF.LStricmp(name, "Line") )       tmpc->line = FSF.atoi(value);
  else if ( !FSF.LStricmp(name, "File") )       tmpc->fileMatch = FSF.atoi(value);
  else if ( !FSF.LStricmp(name, "Col") )
  {
    char *v = (char*)value;
    if ( *value == '?' )
    {
      v++;
      tmpc->searchCol = true;
    }
    tmpc->col = FSF.atoi(v);
  }
}

static HANDLE hReloadCheck = 0, hEvent = 0;
bool reloadNeeded = false;
bool reloadInProcess = false;

DWORD WINAPI reloadCheck(LPVOID)
{
  HANDLE hIdle[2];
  char path[NM];
  char *p = lstrrchr(lstrcpy(path, Info.ModuleName), '\\');
  if ( p )
    *p = 0;
  hIdle[0] = FindFirstChangeNotification(path, true, FILE_NOTIFY_CHANGE_LAST_WRITE);
  if ( !hIdle[0] || hIdle[0] == INVALID_HANDLE_VALUE)
    return 0;
  hIdle[1] = hEvent;
  while ( WaitForMultipleObjects(2, hIdle, false, INFINITE) == WAIT_OBJECT_0 )
  {
    FindNextChangeNotification(hIdle[0]);
    if ( pluginStop || reloadInProcess )
      continue;
    reloadNeeded = true;
  };
  FindCloseChangeNotification(hIdle[0]);
  return 0;
};

static void initThread(void)
{
  DWORD id;
  if ( !hEvent )
    hEvent = CreateEvent(NULL, false, false, NULL);
  hReloadCheck = CreateThread(NULL, 0, reloadCheck, 0, 0, &id);
}

static void doneThread(void)
{
  if ( hReloadCheck )
  {
    SetEvent(hEvent);
    WaitForSingleObject(hReloadCheck, INFINITE);
    CloseHandle(hReloadCheck);
    hReloadCheck = 0;
  }
  if ( hEvent )
  {
    CloseHandle(hEvent);
    hEvent = 0;
  }
}

static void findSectionInXML(char*&p)
{
  bool group;
  char name[MAX_STR_LEN];
  while ( ( getItem(p, name, group) ) != NULL )
    if ( group && !FSF.LStricmp(name, "TrueTpl") )
      break;
}

static void InitMacro()
{
  reloadInProcess = true;
  HANDLE hScreen=Info.SaveScreen(0,0,-1,-1);
  const char *MsgItems[]= { GetMsg(MTitle), GetMsg(MLoading) };
  Info.Message(Info.ModuleNumber,0,NULL,MsgItems,dimOf(MsgItems),0);
  char path[NM];
  addExt(lstrcpy(path, Info.ModuleName), ".xml");
  char *file = lstrrchr(path, '\\');
  *(file++) = 0;
  char *fileBuff = getFile(path, file);
  if ( fileBuff )
  {
    char *item, *p = fileBuff;
    char name[MAX_STR_LEN], value[MAX_STR_LEN];
    TLang *lng = NULL;
    bool group;
    TCollection dummyLang;
    memcpy(&langColl, &dummyLang, sizeof(TLangCollection));
    findSectionInXML(p);
    while ( ( item = getItem(p, name, group) ) != NULL )
    {
      if ( group && !FSF.LStricmp(name, "/True-Tpl") )
        findSectionInXML(p);
      else if ( !group && !FSF.LStricmp(name, "Include") )
      {
        char incFile[NM];
        *incFile = 0;
        TCollection *dc = ( lng == NULL ) ? NULL : &(lng->defineColl);
        while ( parseItem(dc, item, name, value) )
          if ( !FSF.LStricmp(name, "File") )
            lstrcpy(incFile, value);
        if ( *incFile )
        {
          char *incBuff = getFile(path, incFile);
          if ( incBuff )
          {
            char *newBuff = new char[lstrlen(incBuff)+lstrlen(p)+16];
            if ( newBuff )
            {
              lstrcat(lstrcat(lstrcpy(newBuff, incBuff), "\n<TrueTpl>\n"), p);
              delete [] fileBuff;
              fileBuff = p = newBuff;
              findSectionInXML(p);
            }
            delete [] incBuff;
          }
        }
      }
      else if ( group && !FSF.LStricmp(name, "/Language") )
        lng = NULL;
      else if ( group && !FSF.LStricmp(name, "Language") )
      {
        lng = new TLang;
        lng->ignoreCase = false;
//        lng->panel = false;
        lng->mask[0] = lng->desc[0] = lng->imm[0] = lng->immExp[0] = 0;
        lng->defExec.defaults();
        while ( parseItem(NULL, item, name, value) )
        {
          if      ( !FSF.LStricmp(name, "File") )
            lstrcpy(lng->mask, value);
          else if ( !FSF.LStricmp(name, "Desc") )
            lstrcpy(lng->desc, value);
          else if ( !FSF.LStricmp(name, "IgnoreCase") )
            lng->ignoreCase = FSF.atoi(value) ? true : false;
//          else if ( !FSF.LStricmp(name, "Panel") )
//            lng->panel = FSF.atoi(value) ? true : false;
          else
            parseExec(&(lng->defExec), name, value);
        }
        if ( lng->mask[0] )
          langColl.insert(lng);
        else
        {
          delete lng;
          lng = NULL;
        }
      }
      else if ( lng != NULL )
      {
        if ( !group && !FSF.LStricmp(name, "Define") )
        {
          TDefine *tmpd = new TDefine("", "");
          while ( parseItem(&(lng->defineColl), item, name, value) )
            if      ( !FSF.LStricmp(name, "Name") )
              lstrcpy(tmpd->name, value);
            else if ( !FSF.LStricmp(name, "Value") )
              lstrcpy(tmpd->value, value);
          if ( tmpd->name[0] )
            lng->defineColl.insert(tmpd);
          else
            delete tmpd;
        }
        else if ( !group && !FSF.LStricmp(name, "Expand") )
        {
          TMacro *tmpm = new TMacro;
          lstrcpy(tmpm->At, expOnBlank);
          tmpm->immChar = 0;
          tmpm->Word[0] = tmpm->before[0] = tmpm->after[0] = tmpm->MacroText[0] = 0;
          tmpm->FARKey = -1;
          tmpm->atStartup = false;
          int validTag = false;
          while ( parseItem(&(lng->defineColl), item, name, value) )
          {
            if      ( !FSF.LStricmp(name, "Pattern") )
            {
              lstrcpy(tmpm->Word, value);
              if ( tmpm->Word[0] )
              {
                validTag = true;
                char *opDelimiter = lstrchr(tmpm->Word, '|');
                if ( opDelimiter )
                {
                  char rest[MAX_STR_LEN], cc[3] = "(.";
                  *opDelimiter = 0;
                  lstrcpy(rest, opDelimiter+1);
                  for ( char *p = rest ; *p ; p++ )
                  {
                    cc[1] = *p;
                    lstrcat(tmpm->Word, cc);
                  }
                  for ( char *p = rest ; *p ; p++ )
                    lstrcat(tmpm->Word, ")?");
                }
              }
            }
            else if ( !FSF.LStricmp(name, "To") )
              lstrcpy(tmpm->MacroText, value);
            else if ( !FSF.LStricmp(name, "Key") )
            {
              validTag = true;
              tmpm->FARKey = FSF.FarNameToKey(value);
            }
            else if ( !FSF.LStricmp(name, "Imm") )
              tmpm->immChar = *value;
            else if ( !FSF.LStricmp(name, "Init") )
            {
              validTag = true;
              tmpm->atStartup = FSF.atoi(value) ? true : false;
            }
            else if ( !FSF.LStricmp(name, "At") )
            {
              if      ( !FSF.LStricmp(value, "AnyWhere") ) lstrcpy(tmpm->At, expAnyWhere);
              else if ( !FSF.LStricmp(value, "Start" ) )   lstrcpy(tmpm->At, expAtStart);
              else if ( !FSF.LStricmp(value, "Middle") )   lstrcpy(tmpm->At, expAtMiddle);
              else if ( !FSF.LStricmp(value, "End") )      lstrcpy(tmpm->At, expAtEnd);
              else if ( !FSF.LStricmp(value, "Blank") )    lstrcpy(tmpm->At, expOnBlank);
              else                                         lstrcpy(tmpm->At, value);
            }
          }
          if ( validTag )
          {
            if ( tmpm->At[0] )
            {
              char *opDelimiter = lstrstr(tmpm->At, "\\p");
              if ( opDelimiter )
              {
                *opDelimiter = 0;
                lstrcpy(tmpm->before, tmpm->At);
                lstrcpy(tmpm->after, opDelimiter+2);
              }
            }
            if ( tmpm->immChar )
            {
              if ( !lstrchr(lng->immExp, tmpm->immChar) )
              {
                int len = lstrlen(lng->immExp);
                lng->immExp[len++] = tmpm->immChar;
                lng->immExp[len] = 0;
              }
            }
            lng->macroColl.insert(tmpm);
          }
          else
            delete tmpm;
        }
        else if ( !group && !FSF.LStricmp(name, "Comment") )
        {
          TComment *tmpc = new TComment;
          tmpc->mask[0] = 0;
          while ( parseItem(&(lng->defineColl), item, name, value) )
            if ( !FSF.LStricmp(name, "Pattern") )
              lstrcpy(tmpc->mask, value);
          if ( tmpc->mask[0] )
            lng->commentColl.insert(tmpc);
          else
            delete tmpc;
        }
        else if ( !group && !FSF.LStricmp(name, "Exec") )
        {
          TExec *tmpe = new TExec(lng->defExec);
          while ( parseItem(&(lng->defineColl), item, name, value) )
            if ( !FSF.LStricmp(name, "Command") )
              lstrcpy(tmpe->cmd, value);
            else
              parseExec(tmpe, name, value);
            if ( !tmpe->title[0] )
              lstrcpy(tmpe->title, tmpe->cmd);
            if ( !tmpe->title[0] )
              tmpe->cmd[0] = 0;
            lng->execColl.insert(tmpe);
        }
        else if ( !group && !FSF.LStricmp(name, "Compiler") )
        {
          TCompiler *tmpc = new TCompiler();
          while ( parseItem(&(lng->defineColl), item, name, value) )
            if ( !FSF.LStricmp(name, "Name") )
              lstrcpy(tmpc->title, value);
            else
              parseCompiler(tmpc, name, value);
          if ( tmpc->title[0] )
            lng->compilerColl.insert(tmpc);
          else
            delete tmpc;
        }
        else if ( !group && !FSF.LStricmp(name, "Indent") )
        {
          TIndent *tmpi = new TIndent;
          tmpi->mask[0] = tmpi->relative[0] = 0;
          tmpi->BracketsMode = tmpi->start = tmpi->immChar = 0;
          for ( int i = 0 ; i < 2 ; i++ )
            tmpi->indent[i] = 0xFFFF;
          while ( parseItem(&(lng->defineColl), item, name, value) )
          {
            if      ( !FSF.LStricmp(name, "Pattern") )  lstrcpy(tmpi->mask, value);
            else if ( !FSF.LStricmp(name, "Line") )     tmpi->indent[0] = FSF.atoi(value);
            else if ( !FSF.LStricmp(name, "Next") )     tmpi->indent[1] = FSF.atoi(value);
            else if ( !FSF.LStricmp(name, "Imm") )      tmpi->immChar = *value;
            else if ( !FSF.LStricmp(name, "Start") )    tmpi->start = FSF.atoi(value);
            else if ( !FSF.LStricmp(name, "Relative") ) lstrcpy(tmpi->relative, value);
          }
          if ( tmpi->mask[0] )
          {
            if ( tmpi->immChar )
            {
              if ( !lstrchr(lng->imm, tmpi->immChar) )
              {
                int len = lstrlen(lng->imm);
                lng->imm[len++] = tmpi->immChar;
                lng->imm[len] = 0;
              }
            }
            lng->indentColl.insert(tmpi);
          }
          else
            delete tmpi;
        }
        else if ( !group && !FSF.LStricmp(name, "Bracket") )
        {
          TBracket *tmpb = new TBracket;
          TIndent  *tmp0 = new TIndent;
          TIndent  *tmp1 = new TIndent;
          tmpb->open[0] = tmpb->close[0] = 0;
          tmp0->relative[0] = 0;
          tmp0->start = tmp0->immChar = 0;
          tmp0->BracketsMode = 0;
          tmp1->relative[0] = 0;
          tmp1->start = tmp1->immChar = 0;
          tmp1->BracketsMode = 1;
          tmp1->bracketLink = lng->bracketColl.getCount();
          tmp0->indent[0] = 0xFFFF;
          tmp0->indent[1] = 1;
          tmp1->indent[0] = 0;
          tmp1->indent[1] = 0xFFFF;
          while ( parseItem(&(lng->defineColl), item, name, value) )
          {
            if      ( !FSF.LStricmp(name, "Pattern") )  lstrcpy(tmp0->mask, lstrcpy(tmpb->open, value));
            else if ( !FSF.LStricmp(name, "Match") )    lstrcpy(tmp1->mask, lstrcpy(tmpb->close, value));
            else if ( !FSF.LStricmp(name, "Line") )     tmp0->indent[0] = FSF.atoi(value);
            else if ( !FSF.LStricmp(name, "Shift") )    tmp0->indent[1] = FSF.atoi(value);
            else if ( !FSF.LStricmp(name, "Next") )     tmp1->indent[1] = FSF.atoi(value);
            else if ( !FSF.LStricmp(name, "ImmOpen") )  tmp0->immChar = *value;
            else if ( !FSF.LStricmp(name, "ImmClose") ) tmp1->immChar = *value;
            else if ( !FSF.LStricmp(name, "Start") )    tmp0->start = FSF.atoi(value);
          }
          if ( tmpb->open[0] && tmpb->close[0] )
          {
            if ( tmp0->immChar )
            {
              if ( !lstrchr(lng->imm, tmp0->immChar) )
              {
                int len = lstrlen(lng->imm);
                lng->imm[len++] = tmp0->immChar;
                lng->imm[len] = 0;
              }
            }
            if ( tmp1->immChar )
            {
              if ( !lstrchr(lng->imm, tmp1->immChar) )
              {
                int len = lstrlen(lng->imm);
                lng->imm[len++] = tmp1->immChar;
                lng->imm[len] = 0;
              }
            }
            lng->bracketColl.insert(tmpb);
            lng->indentColl.insert(tmp0);
            lng->indentColl.insert(tmp1);
          }
          else
          {
            delete tmpb;
            delete tmp0;
            delete tmp1;
          }
        }
      }
    }
    delete [] fileBuff;
  }
  Info.RestoreScreen(hScreen);
  reloadInProcess = reloadNeeded = false;
  doneThread();
  initThread();
}

static void DoneMacro()
{
  langColl.removeAll();
  doneThread();
}

static inline char *StrLower(char *str)
{
  FSF.LStrlwr(str);
  return str;
}

static int strMatch(char *s, char *pattern, char *prefix, char *suffix, int nb, int bounds[][2] = NULL, int regn[] = NULL)
{
  int ret = false;
  char slashPattern[320];
  static CRegExp reg;
  SMatches m;
  lstrcat(lstrcat(lstrcpy(slashPattern, prefix), pattern), suffix);
  if ( reg.SetExpr(slashPattern) )
  {
    if ( slashPattern[1] == '^' )
      reg.SetNoMoves(true);
    ret = reg.Parse(s, &m) ? true : false;
    if ( ret )
    {
      for ( int i = 0 ; i < nb ; i++ )
      {
        bounds[i][0] = m.s[regn[i]];
        bounds[i][1] = m.e[regn[i]];
      }
    }
  }
  return ret;
}

static TMacro *FindMacro(TLang *lng, char *before, char *after, char expChar, int *delCount)
{
  char *regEnd = lng->ignoreCase ? "\\s*$/i" : "\\s*$/";
  for ( unsigned i = 0 ; i < lng->macroColl.getCount() ; i++ )
  {
    TMacro *mm = (TMacro*)(lng->macroColl[i]);
    if ( expChar && ( expChar != mm->immChar ) )
      continue;
    if ( mm->Word[0] && strMatch(after, mm->after, "/^\\s*", regEnd, 0) )
    {
      char realBefore[MAX_REG_LEN];
      lstrcat(lstrcpy(realBefore, mm->before), mm->Word);
      if ( strMatch(before, realBefore, "/^\\s*", regEnd, 0) )
      {
        if ( delCount )
        {
          regEnd = lng->ignoreCase ? ")$/i" : ")$/";
          int bounds[2] = {0, 0}, zero = 1;
          if ( strMatch(before, mm->Word, "/.*(", regEnd, 1, &bounds, &zero) )
            *delCount = bounds[1]-bounds[0];
          else
            *delCount = 0;
        }
        return mm;
      }
    }
  }
  return NULL;
}

static TMacro *FindMacroKey(TLang *lng, const INPUT_RECORD *Rec)
{
  int rKey = FSF.FarInputRecordToKey(Rec);
  for ( unsigned i = 0 ; i < lng->macroColl.getCount() ; i++ )
  {
    TMacro *mm = (TMacro*)(lng->macroColl[i]);
    if ( mm->FARKey == rKey )
      return mm;
  }
  return NULL;
}

static TMacro *FindStartupMacro(TLang *lng)
{
  for ( unsigned i = 0 ; i < lng->macroColl.getCount() ; i++ )
  {
    TMacro *mm = (TMacro*)(lng->macroColl[i]);
    if ( mm->atStartup )
      return mm;
  }
  return NULL;
}

static TMacro *CheckMacroPos(TLang *lng, TMacro *mm, char *before, char *after)
{
  char *regEnd = lng->ignoreCase ? "\\s*$/i" : "\\s*$/";
  if ( strMatch(after, mm->after, "/^\\s*", regEnd, 0) )
    if ( strMatch(before, mm->before, "/^\\s*", regEnd, 0) )
      return mm;
  return NULL;
}

static char userString[10][80] = { "", "", "", "", "", "", "", "", "", "" };
static char uttlString[10][80] = { "", "", "", "", "", "", "", "", "", "" };
static bool userInp[10];

static void nullUserStrings(void)
{
  for ( int i = 0 ; i < 10 ; i++ )
  {
    *userString[i] = *uttlString[i] = 0;
    userInp[i] = false;
  }
}

static bool insertUserStrings(const char *title = NULL)
{
  static char *h = "TrueTpl";
  unsigned char iX, iY = 0, x0 = 0;
  int uP[10];
  InitDialogItem IBtn[] =
  {
    { DI_TEXT,   5, 1, 0,255, 0, 0,               NULL           },
    { DI_BUTTON, 0, 2, 0,  0, 0, DIF_CENTERGROUP, (char*)MOK     },
    { DI_BUTTON, 0, 2, 0,  0, 0, DIF_CENTERGROUP, (char*)MCancel }
  };
  for ( int i = 0 ; i < 10 ; i++ )
    if ( userInp[i] )
    {
      uP[i] = iY*2+2;
      iY++;
      unsigned char len = (unsigned char)lstrlen(uttlString[i]);
      if ( len > x0 )
        x0 = len;
    }
  iX = (unsigned char)(x0+28);
  unsigned char jX = (unsigned char)(iX-2);
  unsigned char x1 = (unsigned char)(x0+5);
  unsigned char x2 = (unsigned char)(x1+1);
  char *ttl;
  if ( title )
    ttl = (char*)title;
  else
    ttl = (char*)MParams;
  InitDialogItem Init[24] =
  {
    { DI_DOUBLEBOX, 3,  1,  0, 0, 0,        0,           ttl },
    { DI_TEXT,      5,  2, x1, 0, 0,        0,           ""  },
    { DI_EDIT,     x2,  2, jX, 0, (DWORD)h, DIF_HISTORY, ""  },
    { DI_TEXT,      5,  3, x1, 0, 0,        0,           ""  },
    { DI_EDIT,     x2,  3, jX, 0, (DWORD)h, DIF_HISTORY, ""  },
    { DI_TEXT,      5,  4, x1, 0, 0,        0,           ""  },
    { DI_EDIT,     x2,  4, jX, 0, (DWORD)h, DIF_HISTORY, ""  },
    { DI_TEXT,      5,  5, x1, 0, 0,        0,           ""  },
    { DI_EDIT,     x2,  5, jX, 0, (DWORD)h, DIF_HISTORY, ""  },
    { DI_TEXT,      5,  6, x1, 0, 0,        0,           ""  },
    { DI_EDIT,     x2,  6, jX, 0, (DWORD)h, DIF_HISTORY, ""  },
    { DI_TEXT,      5,  7, x1, 0, 0,        0,           ""  },
    { DI_EDIT,     x2,  7, jX, 0, (DWORD)h, DIF_HISTORY, ""  },
    { DI_TEXT,      5,  8, x1, 0, 0,        0,           ""  },
    { DI_EDIT,     x2,  8, jX, 0, (DWORD)h, DIF_HISTORY, ""  },
    { DI_TEXT,      5,  9, x1, 0, 0,        0,           ""  },
    { DI_EDIT,     x2,  9, jX, 0, (DWORD)h, DIF_HISTORY, ""  },
    { DI_TEXT,      5, 10, x1, 0, 0,        0,           ""  },
    { DI_EDIT,     x2, 10, jX, 0, (DWORD)h, DIF_HISTORY, ""  },
    { DI_TEXT,      5, 11, x1, 0, 0,        0,           ""  },
    { DI_EDIT,     x2, 11, jX, 0, (DWORD)h, DIF_HISTORY, ""  },
  };
  bool res = true;
  if ( iY )
  {
    res = false;
    unsigned char bX = (unsigned char)(iX+1);
    unsigned char bY = (unsigned char)(iY+4);
    Init[0].X2 = bX;
    Init[0].Y2 = bY;
    IBtn[0].Y1 = (unsigned char)(bY-2);
    IBtn[1].Y1 = IBtn[2].Y1 = (unsigned char)(bY-1);
    int size = iY*2+4;
    memcpy(Init+(iY*2+1), &IBtn, sizeof(IBtn));
    iY = 0;
    for ( int i = 0 ; i < 10 ; i++ )
    {
      if ( userInp[i] )
      {
        int it = iY*2;
        Init[it+1].Data = uttlString[i];
        Init[it+2].Data = userString[i];
        iY++;
      }
    }
    int OK = iY*2+2;
    int Result = ExecDialogEx(Init,size,1,OK,NULL,NULL,NULL);
    if ( Result == OK )
      res = true;
  }
  return res;
}

struct TEditorPos
{
  TEditorPos() { Default(); }
  int Row, Col;
  int TopRow,LeftCol;
  void Default(void) { Row = Col = TopRow = LeftCol = -1; }
};

static TEditorPos EditorGetPos(void)
{
  TEditorPos r;
  EditorInfo ei;
  Info.EditorControl(ECTL_GETINFO, &ei);
  r.Row = ei.CurLine;
  r.Col = ei.CurPos;
  r.TopRow = ei.TopScreenLine;
  r.LeftCol = ei.LeftPos;
  return r;
}

static void EditorSetPos(TEditorPos pos)
{
  EditorSetPosition sp;
  sp.CurLine = pos.Row;
  sp.CurPos = pos.Col;
  sp.TopScreenLine = pos.TopRow;
  sp.LeftPos = pos.LeftCol;
  sp.CurTabPos = sp.Overtype = -1;
  Info.EditorControl(ECTL_SETPOSITION,&sp);
}

void EditorSetPos(int line, int col, int topline = -1, int leftcol = -1)
{
  EditorSetPosition sp;
  sp.CurLine = ( line < -1 ) ? -1 : line;
  sp.CurPos = ( col < -1 ) ? -1 : col;
  sp.TopScreenLine = topline;
  sp.LeftPos = leftcol;
  sp.CurTabPos = sp.Overtype = -1;
  Info.EditorControl(ECTL_SETPOSITION, &sp);
}

static char* Editor2OEM(char *str)
{
  EditorConvertText ect;
  ect.Text = str;
  ect.TextLength = lstrlen(str);
  Info.EditorControl(ECTL_EDITORTOOEM, &ect);
  return str;
}

static void EditorGetStr(EditorGetString *gs, int line = -1)
{
  gs->StringNumber = line;
  Info.EditorControl(ECTL_GETSTRING, gs);
}

static void EditorSetStr(char *src, int line = -1)
{
  EditorSetString st;
  st.StringNumber = line;
  st.StringText = src;
  st.StringEOL = 0;
  st.StringLength = lstrlen(src);
  Info.EditorControl(ECTL_SETSTRING, &st);
}

void EditorSetPosEx(int line, int initCol, const char *colSearch)
{
  int col = ( initCol < 0 ) ? -1 : initCol;
  EditorSetPos(line, col);
  TEditorPos pos = EditorGetPos();
  if ( pos.Row == pos.TopRow )
  {
    int tl = line-5;
    if ( tl < 1 )
      tl = 1;
    EditorSetPos(line, -1, tl);
  }
  EditorGetString gs;
  EditorGetStr(&gs);
  if ( ( initCol == -2 ) && *colSearch )
  {
    char s[MAX_STR_LEN];
    lstrcpyn(s, gs.StringText, gs.StringLength+1);
    char *found = lstrstr(Editor2OEM(s), colSearch);
    if ( found )
    {
      int newCol = found-s;
      if ( newCol != col )
        EditorSetPos(-1, newCol);
    }
  }
  pos = EditorGetPos();
  if ( pos.Col > gs.StringLength )
    EditorSetPos(-1, gs.StringLength);
}

static void EditorProcessKey(char ascii)
{
  INPUT_RECORD tr;
  tr.EventType = KEY_EVENT;
  tr.Event.KeyEvent.bKeyDown = true;
  tr.Event.KeyEvent.wRepeatCount = 1;
  tr.Event.KeyEvent.wVirtualKeyCode = 0;
  tr.Event.KeyEvent.uChar.AsciiChar = ascii;
  tr.Event.KeyEvent.dwControlKeyState = 0;
  Info.EditorControl(ECTL_PROCESSINPUT, &tr);
}

static inline void EditorProcessFARKey(int key)
{
  Info.EditorControl(ECTL_PROCESSKEY, (void*)key);
}

static inline void EditorInsertText(char *s)
{
  Info.EditorControl(ECTL_INSERTTEXT, (void*)s);
}

static inline void redraw(void)
{
  Info.EditorControl(ECTL_REDRAW, 0);
}

static void EditorProcessReturnKey(int before = -1, int after = -1)
{
  INPUT_RECORD tr;
  tr.EventType = KEY_EVENT;
  tr.Event.KeyEvent.bKeyDown = true;
  tr.Event.KeyEvent.wRepeatCount = 1;
  tr.Event.KeyEvent.wVirtualKeyCode = VK_RETURN;
  tr.Event.KeyEvent.dwControlKeyState = 0;
  if ( before != -1 )
    pluginBusy = before;
  Info.EditorControl(ECTL_PROCESSINPUT, &tr);
  if ( after != -1 )
    pluginBusy = after;
}

static inline void insIns(char *ins, int& insPos)
{
  if ( *ins )
  {
    EditorInsertText(ins);
    insPos = *ins = 0;
    redraw();
  }
}

static void selectBlock(int X1, int Y1, int X2, int Y2)
{
  EditorSelect es;
  es.BlockType = BTYPE_STREAM;
  es.BlockStartLine = min(Y2, Y1);
  es.BlockStartPos = min(X1, X2);
  es.BlockHeight = max(Y1, Y2)-min(Y1, Y2)+1;
  es.BlockWidth = max(X1, X2)-min(X1, X2);
  Info.EditorControl(ECTL_SELECT, (void*)&es);
}

static bool scanUserInput(bool inMacro, char macro, const char *MacroText, const char *title = NULL)
{
  char titleStr[80] = "";
  nullUserStrings();
  for ( char *p = (char*)MacroText ; *p ; p++ )
  {
    char *cb;
    if ( *p == macro )
    {
      ++p;
      if ( lstrchr("0123456789", *p) )
      {
        int userStrN = *p-'0';
        if ( !userInp[userStrN] )
          lstrcpy(uttlString[userStrN], titleStr);
        userInp[userStrN] = true;
        if ( p[1] == '\'' )
        {
          p++;
          *titleStr = 0;
          char *s = titleStr;
          while ( *++p && *p != '\'' )
            *s++ = *p;
          *s = 0;
          lstrcpy(userString[userStrN], titleStr);
        }
      }
      else if ( *p == '?' )
      {
        if ( p[1] == '\'' )
        {
          p++;
          *titleStr = 0;
          char *s = titleStr;
          while ( *++p && *p != '\'' )
            *s++ = *p;
          *s = 0;
        }
      }
      else if ( inMacro )
      {
        switch ( *p )
        {
          case '!':
            ++p;
            break;
          case '{':
            if ( ( cb = lstrchr(p+1, '}') ) != NULL )
              p = cb;
            break;
          case 'w':
          case 'd':
            if ( p[1] == '\'' )
            {
              p++;
              while ( *++p && *p != '\'' )
                ;
            }
            break;
          default:
            break;
        }
      }
    }
  }
  return insertUserStrings(title);
}

static void RunMacro(TMacro *m)
{
  if ( scanUserInput(true, '\\', m->MacroText) )
  {
    int userStrN, insPos = 0, setPos = 0, setStart = 0, setEnd = 0;
    bool makeTime = false;
    char ins[MAX_STR_LEN] = "";
    TEditorPos pos, curPos, selStart, selEnd;
    pluginBusy = 1;
    for ( char *p = m->MacroText ; *p ; p++ )
    {
      char *cb;
      if ( *p == '\\' )
      {
        insIns(ins, insPos);
        switch ( *++p )
        {
          case 'p':
            setPos = 1;
            pos = EditorGetPos();
            break;
          case 's':
            setStart = 1;
            selStart = EditorGetPos();
            break;
          case 'e':
            setEnd = 1;
            selEnd = EditorGetPos();
            break;
          case 0:
          case '\\':
            EditorProcessFARKey(KEY_BACKSLASH);
            break;
          case 't':
            EditorProcessFARKey(KEY_TAB);
            break;
          case 'b':
            EditorProcessFARKey(KEY_BS);
            break;
          case '^':
            EditorProcessFARKey(KEY_UP);
            break;
          case 'v':
            EditorProcessFARKey(KEY_DOWN);
            break;
          case '<':
            EditorProcessFARKey(KEY_LEFT);
            break;
          case '>':
            EditorProcessFARKey(KEY_RIGHT);
            break;
          case 'h':
          case '[':
            EditorProcessFARKey(KEY_HOME);
            break;
          case ']':
            EditorProcessFARKey(KEY_END);
            break;
          case '(':
            EditorProcessFARKey(KEY_LEFT|KEY_CTRL);
            break;
          case ')':
            EditorProcessFARKey(KEY_RIGHT|KEY_CTRL);
            break;
          case 'r':
            EditorProcessReturnKey(0, 1);
            break;
          case 'n':
            EditorProcessReturnKey();
            break;
          case '!':
            pluginBusy = 0;
            EditorProcessKey(*++p);
            pluginBusy = 1;
          case '?':
            if ( p[1] == '\'' )
            {
              p++;
              while ( *++p && *p != '\'' )
                ;
            }
            break;
          case '0': case '1': case '2': case '3': case '4':
          case '5': case '6': case '7': case '8': case '9':
            userStrN = *p-'0';
            if ( p[1] == '\'' )
            {
              p++;
              while ( *++p && *p != '\'' )
                ;
            }
            Info.EditorControl(ECTL_INSERTTEXT, userString[userStrN]);
            break;
          case '{':
            if ( ( cb = lstrchr(p+1, '}') ) == NULL )
              EditorProcessKey(*p);
            else
            {
              char ch = *cb;
              *cb = 0;
              EditorProcessFARKey(FSF.FarNameToKey(p+1));
              *cb = ch;
              p = cb;
            }
            break;
          case 'w':
            makeTime = true;
          case 'd':
          {
            char strD[80], format[80] = "", *fmt = NULL;
            SYSTEMTIME SysDate;
            GetLocalTime(&SysDate);
            if ( p[1] == '\'' )
            {
              char *t = format;
              p++;
              while ( *++p && *p != '\'' )
                *t++ = ( *p == '`' ) ? '\'' : *p;
              *t = 0;
              fmt = format;
            }
            ( makeTime ? GetTimeFormat : GetDateFormat )(LOCALE_SYSTEM_DEFAULT, 0, &SysDate, fmt, strD, sizeof(strD));
            CharToOem(strD, strD);
            EditorInsertText(strD);
            break;
          }
          default:
            pluginBusy = 0;
            EditorProcessKey(*p);
            pluginBusy = 1;
            break;
        }
      }
      else
      {
        ins[insPos++] = *p;
        ins[insPos] = 0;
      }
    }
    insIns(ins, insPos);
    pluginBusy = 0;
    if ( setPos )
      EditorSetPos(pos);
    if ( setStart && setEnd )
      selectBlock(selStart.Col, selStart.Row, selEnd.Col, selEnd.Row);
  }
}

int WINAPI _export GetMinFarVersion(void)
{
  /*return FARMANAGERVERSION;*/
  return MAKEFARVERSION(1,70,1282);
}

void WINAPI _export SetStartupInfo(const PluginStartupInfo *Info)
{
  ::Info = *Info;
  IsOldFar = true;
  if ( Info->StructSize >= sizeof(PluginStartupInfo) )
  {
    ::FSF = *Info->FSF;
    ::Info.FSF = &::FSF;
    IsOldFar = false;
    FSF.sprintf(PluginRootKey,"%s\\TRUE-TPL",Info->RootKey);
    GetRegKey(HKEY_CURRENT_USER, "", "Disable", pluginStop, 0);
    GetRegKey(HKEY_CURRENT_USER, "", "UseScrollLock", scrollStop, 0);
    GetRegKey(HKEY_CURRENT_USER, "", "Key", defExpandFKey, KEY_SPACE);
    InitMacro();
  }
}

TEICollection *eList = NULL;

static void initEList(void)
{
  if ( eList == NULL )
    eList = new TEICollection;
}

void WINAPI _export GetPluginInfo(PluginInfo *Info)
{
  if ( !IsOldFar )
  {
    static const char *PluginMenuStrings[1];
    Info->StructSize = sizeof(*Info);
    Info->Flags = PF_EDITOR|PF_VIEWER|PF_DISABLEPANELS;
    Info->DiskMenuStringsNumber = 0;
    Info->PluginConfigStringsNumber = 0;
    PluginMenuStrings[0] = GetMsg(MTitle);
    Info->PluginMenuStrings = PluginMenuStrings;
    Info->PluginMenuStringsNumber = 1;
    static const char *PluginCfgStrings[1];
    PluginCfgStrings[0] = GetMsg(MTitle);
    Info->PluginConfigStrings=PluginCfgStrings;
    Info->PluginConfigStringsNumber=dimOf(PluginCfgStrings);
    static const char *PluginPrefix = GetMsg(MPrefix);
    Info->CommandPrefix = PluginPrefix;
  }
}

static int newExpandFKey = defExpandFKey;

long WINAPI KeyPressDlgProc(HANDLE hDlg, int Msg, int Param1, long Param2)
{
  char KeyText[50];
  static int LastKey;
  static DWORD *KMParam = NULL;
  bool gotoM1 = false;
  if ( Msg == DN_INITDIALOG )
  {
    KMParam = (DWORD*)Param2;
    LastKey = 0;
    static const char * const PreDefKeyName[] = { "CtrlDown" };
    for ( int i = 0 ; i < dimOf(PreDefKeyName) ; i++ )
      Info.SendDlgMessage(hDlg,DM_LISTADDSTR,2,(long)PreDefKeyName[i]);
    FSF.FarKeyToName(*KMParam,KeyText,0);
    Info.SendDlgMessage(hDlg,DM_SETTEXTPTR,2,(long)KeyText);
  }
  else if ( Param1 == 2 && Msg == DN_EDITCHANGE )
  {
    LastKey = 0;
    Param2 = FSF.FarNameToKey(((FarDialogItem*)Param2)->Data);
    if ( Param2 != -1 )
      gotoM1 = true;
  }
  else if ( Msg == DN_KEY )
  {
    if ( Param2 == KEY_ESC || Param2 == KEY_ENTER && LastKey == KEY_ENTER || Param2 == KEY_CTRLDOWN || Param2 == KEY_F1 )
      return FALSE;
    if ( Param2 == KEY_ENTER && LastKey && LastKey != KEY_ENTER )
      return FALSE;
    gotoM1 = true;
  }
  else if ( Msg == DN_CTLCOLORDLGITEM )
  {
    Param2 &= 0xFF00FFFFU;
    Param2 |= (Param2 & 0xFF) << 16;
    return Param2;
  }
  if ( gotoM1 )
  {
    FSF.FarKeyToName(Param2,KeyText,0);
    DWORD key = (DWORD)Param2;
    int msg = MGetKeyPrompt;
    if ( ( key & ~(KEY_CTRL|KEY_RCTRL|KEY_SHIFT|KEY_ALT|KEY_RALT) ) == 0 )
      msg = MGetKeyError;
    else
      switch ( key )
      {
        case KEY_ALTF5:        case KEY_ALTF9:
        case KEY_ALTF11:       case KEY_ALTSHIFTF9:
        case KEY_CTRLF10:      case KEY_CTRLO:
        case KEY_CTRLSHIFTTAB: case KEY_CTRLTAB:
        case KEY_CTRLW:        case KEY_F10:
        case KEY_F11:          case KEY_F12:
        case KEY_F2:           case KEY_F6:
        case KEY_SHIFTF10:     case KEY_SHIFTF2:
        case KEY_ENTER:
          msg = MGetKeyError;
          break;
      }
    if ( msg == MGetKeyPrompt )
    {
      char KeyName[512];
      HKEY hKey;
      if ( RegOpenKeyEx(HKEY_CURRENT_USER,"Software\\Far\\KeyMacros\\Editor",0,KEY_READ,&hKey) == ERROR_SUCCESS )
      {
        bool cont = true;
        for ( DWORD i = 0 ; cont ; i++ )
        {
          DWORD size = sizeof(KeyName);
          LONG res = RegEnumKeyEx(hKey, i, KeyName, &size, NULL, NULL, NULL, NULL);
          switch ( res )
          {
            case ERROR_SUCCESS:
              if ( !lstrcmp(KeyName, KeyText) )
              {
                msg = MGetKeyMacros;
                cont = false;
              }
              break;
            case ERROR_NO_MORE_ITEMS:
              cont = false;
              break;
            default:
              cont = false;
              break;
          }
        }
      }
      RegCloseKey(hKey);
    }
    Info.SendDlgMessage(hDlg,DM_SETTEXTPTR,1,(long)GetMsg(msg));
    *KMParam = key;
    Info.SendDlgMessage(hDlg,DM_SETTEXTPTR,2,(long)KeyText);
    Info.SendDlgMessage(hDlg,DM_REDRAW,0,0);
    LastKey = Param2;
    return TRUE;
  }
  return Info.DefDlgProc(hDlg,Msg,Param1,Param2);
}

#define CFG_KEYBUTTON 4
#define CFG_KEYPROMPT 5

long WINAPI ConfigDlgProc(HANDLE hDlg, int Msg, int Param1, long Param2)
{
  char KeyText[50];
  switch ( Msg )
  {
    case DN_BTNCLICK:
      if ( Param1 == CFG_KEYBUTTON )
      {
        FSF.FarKeyToName(newExpandFKey,KeyText,0);
        InitDialogItem InitItems[] =
        {
          { DI_DOUBLEBOX, 3,1,40,4,0,0, (char*)MKey          },
          { DI_TEXT,      5,2, 0,0,0,0, (char*)MGetKeyPrompt },
          { DI_COMBOBOX,  5,3,38,3,1,0, KeyText              }
        };
        DWORD newKey = newExpandFKey;
        int ExitCode = ExecDialogEx(InitItems,dimOf(InitItems),2,2,NULL,KeyPressDlgProc,(long)&newKey);
        if ( ExitCode != -1 )
        {
          FSF.FarKeyToName(newExpandFKey = newKey,KeyText,0);
          Info.SendDlgMessage(hDlg,DM_SETTEXTPTR,CFG_KEYPROMPT,(long)&KeyText);
          Info.SendDlgMessage(hDlg,DM_REDRAW,0,0);
        }
        return TRUE;
      }
      break;
  }
  return Info.DefDlgProc(hDlg,Msg,Param1,Param2);
}

int Config(void)
{
  char KeyText[50];
  FSF.FarKeyToName(newExpandFKey = defExpandFKey,KeyText,0);
  InitDialogItem InitItems[] =
  {
    { DI_DOUBLEBOX, 3,1,55,  8,0, 0,               (char*)MTitle   },
    { DI_CHECKBOX,  5,2, 0,  0,pluginStop,0,       (char*)MDisable }, // 1
    { DI_CHECKBOX,  5,3, 0,  0,scrollStop,0,       (char*)MPause   }, // 2
    { DI_TEXT,      5,4, 0,255,0, 0,               NULL            },
    { DI_BUTTON,    5,5, 0,  0,0, 0,               (char*)MKey     }, // 4
    { DI_TEXT,     30,5, 0,  0,0, 0,               KeyText         }, // 5
    { DI_TEXT,      5,6, 0,255,0, 0,               NULL            },
    { DI_BUTTON,    0,7, 0,  0,0, DIF_CENTERGROUP, (char*)MReload  }, // 7 OK
    { DI_BUTTON,    0,7, 0,  0,0, DIF_CENTERGROUP, (char*)MCancel  }
  };
  const OK = 7;
  int ExitCode = ExecDialogEx(InitItems,dimOf(InitItems),1,OK,NULL,ConfigDlgProc,NULL);
  if ( ExitCode == OK )
  {
    SetRegKey(HKEY_CURRENT_USER, "", "Disable", pluginStop = InitItems[1].Selected);
    SetRegKey(HKEY_CURRENT_USER, "", "UseScrollLock", scrollStop = InitItems[2].Selected);
    SetRegKey(HKEY_CURRENT_USER, "", "Key", defExpandFKey = newExpandFKey);
    DoneMacro();
    InitMacro();
    return true;
  }
  return false;
}

int WINAPI _export Configure(int ItemNumber)
{
  if ( ItemNumber )
    return false;
  return Config();
}

static inline int matched(const char *mask, const char *value)
{
  return FSF.ProcessName(mask, (char*)value, PN_CMPNAMELIST|PN_SKIPPATH);
}

static char *FirstNonSpace(const char *s)
{
  char *i = (char*)s;
  while ( *i )
  {
    if ( !isCharSpace(*i) )
      return i;
    i++;
  }
  return NULL;
}

static int FindIndent(char c, TLang *lng, char *str, int foundBounds[2])
{
  char line[MAX_STR_LEN];
  int bounds[2];
  for ( unsigned j = 0 ; j < lng->indentColl.getCount() ; j++ )
  {
    TIndent *tmpi = (TIndent*)(lng->indentColl[j]);
    lstrcpy(line, tmpi->mask);
    char ic = ((TIndent*)(lng->indentColl[j]))->immChar;
    if ( lng->ignoreCase )
      ic = (char)FSF.LLower(ic);
    if ( ( c == '\0' ) || ( c == ic ) )
      if ( strMatch(str, line, "/^", lng->ignoreCase ? "$/i" : "$/", 1, &bounds, &tmpi->start) )
      {
        foundBounds[0] = bounds[0];
        foundBounds[1] = bounds[1];
        return j;
      }
  }
  return -1;
}

static int FindBrackets(TLang *lng, char *str, char **foundBrackets)
{
  char line[MAX_STR_LEN], *closeReg = "$/";
  if ( lng->ignoreCase )
  {
    StrLower(line);
    closeReg = "$/i";
  }
  int bounds[2], foundBracketsNum = 0;
  for ( unsigned j = 0 ; j < lng->indentColl.getCount() ; j++ )
  {
    TIndent *tmpi = (TIndent*)(lng->indentColl[j]);
    lstrcpy(line, tmpi->mask);
    if ( strMatch(str, line, "/^", closeReg, 1, &bounds, &tmpi->start) )
    {
      if ( tmpi->indent[0] != 0xFFFF && tmpi->BracketsMode )
      {
        TBracket *br = (TBracket*)(lng->bracketColl[tmpi->bracketLink]);
        foundBrackets[foundBracketsNum++] = br->open;
      }
    }
  }
  return foundBracketsNum;
}

static char *LookForOpenBracket(TEditorPos& p, TLang *lng, char **openBr, int nOpenBr)
{
  static char *c, buff[MAX_STR_LEN];
  static EditorGetString gs;
  int ret = -1;
  char *closeReg = lng->ignoreCase ? "$/i" : "$/";
  while ( --p.Row >= 0 )
  {
    EditorSetPos(p);
    EditorGetStr(&gs);
    if ( gs.StringLength )
    {
      if ( ( c = FirstNonSpace(gs.StringText) ) != NULL )
      {
        Editor2OEM(FSF.Trim(lstrcpy(buff, c)));
        if ( lng->ignoreCase )
          StrLower(buff);
        for ( int i = 0 ; i < nOpenBr ; i++ )
        {
          if ( strMatch(buff, openBr[i], "/^", closeReg, NULL, 0) )
          {
            ret = c-gs.StringText;
            break;
          }
        }
        if ( ret != -1 )
          break;
        unsigned n = lng->bracketColl.getCount();
        for ( unsigned i = 0 ; i < n ; i++ )
        {
          TBracket *br = (TBracket*)(lng->bracketColl[i]);
          if ( strMatch(buff, br->close, "/^", closeReg, NULL, 0 ) )
          {
            char *result = NULL, **br4look = new char*[n];
            int nbr4look = FindBrackets(lng, buff, br4look);
            if ( nbr4look )
              result = LookForOpenBracket(p, lng, br4look, nbr4look);
            delete [] br4look;
            if ( result )
              break;
            return NULL;
          }
        }
      }
    }
  }
  if ( ret != -1 )
  {
    memcpy(buff, gs.StringText, ret);
    buff[ret] = 0;
    return buff;
  }
  return NULL;
}

static char *FindOpenBracket(TLang *lng, char **openBr, int n)
{
  TEditorPos p, o = EditorGetPos();
  p.Default();
  p.Row = o.Row;
  char *ret = LookForOpenBracket(p, lng, openBr, n);
  EditorSetPos(o);
  return ret;
}

static char *prevStringIndent(void)
{
  static char buff[MAX_STR_LEN];
  EditorGetString gs;
  int ret = 0;
  TEditorPos p, o = EditorGetPos();
  p.Default();
  p.Row = o.Row;
  while ( --p.Row >= 0 )
  {
    EditorSetPos(p);
    EditorGetStr(&gs);
    if ( gs.StringLength )
    {
      char *c = FirstNonSpace(gs.StringText);
      if ( c )
      {
        ret = c-gs.StringText;
        break;
      }
    }
  }
  EditorSetPos(o);
  memcpy(buff,gs.StringText, ret);
  buff[ret]=0;
  return buff;
}

static int shiftLine(int i)
{
  TEditorPos pos = EditorGetPos();
  pluginBusy = 1;
  if ( i > 0 )
    for ( int j = 0 ; j < i ; j++ )
      EditorProcessFARKey(KEY_TAB);
  else if ( i < 0 )
  {
    EditorInfo ei;
    Info.EditorControl(ECTL_GETINFO, &ei);
    int ind = -i*ei.TabSize;
    for ( int j = 0 ; j < ind ; j++ )
      EditorProcessFARKey(KEY_BS);
  }
  TEditorPos pos2 = EditorGetPos();
  pluginBusy = 0;
  return pos2.Col-pos.Col;
}

static int TryIndent(int ret, TLang *lng, int j, char *str, char *realStr, int oldIndent, int foundBounds[2])
{
  TIndent *tmpi = (TIndent*)(lng->indentColl[j]);
  int iCurr = tmpi->indent[0];
  int iNext = tmpi->indent[1];
  if ( iCurr != 0xFFFF )
  {
    bool processIndent = true;
    char *opencol = NULL;
    TEditorPos pos = EditorGetPos();
    if ( *(tmpi->relative) )
    {
      char *rel = tmpi->relative;
      if ( ( opencol = FindOpenBracket(lng, &rel, 1) ) == NULL )
        processIndent = false;
    }
    else if ( tmpi->BracketsMode )
    {
      unsigned n = lng->bracketColl.getCount();
      char **br4look = new char*[n];
      int nbr4look = FindBrackets(lng, str, br4look);
      processIndent = false;
      if ( nbr4look )
        opencol = FindOpenBracket(lng, br4look, nbr4look);
      if ( opencol )
        processIndent = true;
      delete [] br4look;
    }
    if ( processIndent )
    {
SysLog("process indent\n");
      if ( opencol == NULL )
        opencol = prevStringIndent();
      size_t olen = lstrlen(opencol);
      char *ptr = new char[olen+lstrlen(str)+1];
      lstrcat(lstrcpy(ptr, opencol), realStr);
SysLog("EditorSetStr('%s');\n", ptr);
      EditorSetStr(ptr);
      EditorSetPos(-1, olen);
SysLog("int sh = shiftLine(%d);\n", iCurr);
      int sh = shiftLine(iCurr);
SysLog("EditorSetPos(-1, (pos.Col=%d)+(olen=%d)-(oldIndent=%d)+(sh=%d)=%d)\n", pos.Col, olen, oldIndent, sh, pos.Col+olen-oldIndent+sh);
      EditorSetPos(-1, pos.Col+olen-oldIndent+sh);
SysLog("process indent done\n");
    }
  }
  if ( ret )
  {
    EditorProcessReturnKey(1, 0);
    if ( tmpi->start )
    {
      pluginBusy = 1;
      for ( int k = foundBounds[0] ; k < foundBounds[1] ; k++ )
        EditorProcessKey(realStr[k]);
      pluginBusy = 0;
    }
    if ( iNext != 0xFFFF )
      shiftLine(iNext);
  }
  else
    redraw();
  return IGNORE_EVENT;
}

static int inComment(TLang *lng, const char *realLine, unsigned pos)
{
  int ic = lng->ignoreCase;
  for ( unsigned i = 0 ; i < lng->commentColl.getCount() ; i++ )
  {
    char line[MAX_STR_LEN];
    TComment *tmpc = (TComment*)(lng->commentColl[i]);
    lstrcpy(line, realLine);
    if ( ic )
      StrLower(line);
    int bounds[2], zero = 0;
    if ( strMatch(line, tmpc->mask, "/", ic ? "/i" : "/", 1, &bounds, &zero) )
      if ( pos >= (unsigned)bounds[0] && pos <= (unsigned)bounds[1] )
        return 1;
  }
  return 0;
}

static bool checkMultipleChoice(TLang *lng, TMacro*&mc, char *before, char *after, int len)
{
  const char *menuPrefix = "\\~";
  unsigned pl = lstrlen(menuPrefix);
  if ( (unsigned)lstrlen(mc->MacroText) > pl && !FSF.LStrnicmp(mc->MacroText, menuPrefix, pl) )
  {
    unsigned lc = 0;
    char *p = mc->MacroText, *p1;
    while ( ( p1 = lstrstr(p, menuPrefix) ) != NULL )
    {
      p = p1+pl;
      lc++;
    }
    if ( lc )
    {
      FarMenuItem *amenu = new FarMenuItem[lc];
      TMacro **amacro = new TMacro*[lc];
      if ( amenu )
      {
        unsigned i = 0;
        p = mc->MacroText;
        while ( ( p1 = lstrstr(p, menuPrefix) ) != NULL )
        {
          p = p1+pl;
          char ch, *p2 = lstrstr(p, menuPrefix);
          if ( p2 )
          {
            ch = *p2;
            *p2 = 0;
          }
          TMacro *tm;
          char tmpBefore[MAX_STR_LEN], *p3 = lstrchr(p, '=');
          lstrcpy(tmpBefore, before);
          tmpBefore[lstrlen(tmpBefore)-len] = 0;
          if ( p3 )
          {
            *p3 = 0;
            lstrcat(tmpBefore, p);
            tm = FindMacro(lng, tmpBefore, after, 0, NULL);
            lstrcpy(amenu[i].Text, p3+1);
            *p3 = '=';
          }
          else
          {
            lstrcat(tmpBefore, p);
            tm = FindMacro(lng, tmpBefore, after, 0, NULL);
            lstrcpy(amenu[i].Text, p);
          }
          if ( p2 )
            *p2 = ch;
          if ( tm )
          {
            amacro[i] = tm;
            amenu[i].Selected = amenu[i].Checked = amenu[i].Separator = false;
            i++;
          }
        }
        int res = Info.Menu(Info.ModuleNumber,-1,-1,0,FMENU_WRAPMODE,GetMsg(MSelectMacro),NULL,NULL,NULL,NULL,amenu,i);
        bool retval = false;
        if ( res != -1 && amacro[res] )
        {
          mc = amacro[res];
          retval = true;
        }
        delete [] amenu;
        delete [] amacro;
        return retval;
      }
    }
  }
  return true;
}

static int TryMacro(TLang *lng, EditorGetString& gs, TEditorPos& ep, char *line, char expChar)
{
  if ( expChar || !isCharSpace(gs.StringText[ep.Col-1]) )
  {
    Editor2OEM(lstrcpy(line, gs.StringText));
    if ( !inComment(lng, line, ep.Col) )
    {
      char before[MAX_STR_LEN];
      char *after = line+ep.Col;
      lstrcpy(before, line);
      int len = before[ep.Col] = 0;
      if ( expChar )
      {
        before[ep.Col] = expChar;
        before[ep.Col+1] = 0;
      }
      TMacro *mc = FindMacro(lng, before, after, expChar, &len);
      if ( mc && checkMultipleChoice(lng, mc, before, after, len) )
      {
        if ( expChar )
          len--;
        pluginBusy = 1;
        for ( int i = 0 ; i < len ; i++ )
          EditorProcessFARKey(KEY_BS);
        pluginBusy = 0;
        RunMacro(mc);
        redraw();
        return 1;
      }
    }
  }
  return 0;
}

static bool isFile(const char *s)
{
  WIN32_FIND_DATA ff;
  HANDLE h = FindFirstFile(s, &ff);
  bool res = h != INVALID_HANDLE_VALUE;
  FindClose(h);
  return res;
}

static int findLngID(const char *fn)
{
  for ( unsigned i = 0 ; i < langColl.getCount() ; i++ )
  {
    TLang *lng = (TLang*)(langColl[i]);
    if ( matched(lng->mask, fn) )
      return i;
  }
  return -1;
}

static int eListInsert(EditorInfo *ei)
{
  int newFile = !isFile(ei->FileName);
  return eList->insert(ei->EditorID, findLngID(ei->FileName), ei->FileName, newFile);
}

int WINAPI _export ProcessEditorEvent(int Event, void *Param)
{
  static INPUT_RECORD dummy;
  EditorInfo ei;
  initEList();
  switch ( Event )
  {
    case EE_READ:
      Info.EditorControl(ECTL_GETINFO, &ei);
      eListInsert(&ei);
      memset(&dummy, 0, sizeof(dummy));
      dummy.EventType = KEY_EVENT;
      dummy.Event.KeyEvent.bKeyDown = true;
      Info.EditorControl(ECTL_PROCESSINPUT, &dummy);
      return 0;
    case EE_CLOSE:
      eList->removeID(*((int*)Param));
      return 0;
    case EE_REDRAW:
      if ( reloadNeeded )
      {
        reloadInProcess = true;
        DoneMacro();
        InitMacro();
      }
      return 0;
    default:
      return 0;
  }
}

int WINAPI _export ProcessEditorInput(const INPUT_RECORD *Rec)
{
  if ( pluginStop || pluginBusy || Rec->EventType != KEY_EVENT || Rec->Event.KeyEvent.bKeyDown == 0 )
    return PROCESS_EVENT;
  int vState = Rec->Event.KeyEvent.dwControlKeyState;
  if ( scrollStop && ( vState & SCROLLLOCK_ON ) )
    return PROCESS_EVENT;
  EditorInfo ei;
  Info.EditorControl(ECTL_GETINFO, &ei);
  initEList();
  int edId = eList->findID(ei.EditorID);
  if ( edId == -1 )
    edId = eListInsert(&ei);
  TEInfo *te = (TEInfo*)(*eList)[edId];
  int id = te->lang;
  if ( id != -1 )
  {
    TLang *lng = (TLang*)(langColl[id]);
    if ( lng )
    {
      char line[MAX_STR_LEN];
      EditorGetString gs;
      TEditorPos epos = EditorGetPos();
      TMacro *fm;
      if ( te->newFile )
      {
        te->newFile = 0;
        if ( !Rec->Event.KeyEvent.wVirtualKeyCode )
        {
          if ( ( fm = FindStartupMacro(lng) ) != NULL )
          {
            RunMacro(fm);
            redraw();
          }
          return IGNORE_EVENT;
        }
      }
      if ( ( fm = FindMacroKey(lng, Rec) ) != NULL )
      {
        EditorGetStr(&gs);
        Editor2OEM(lstrcpy(line, gs.StringText));
        if ( !inComment(lng, line, epos.Col) )
        {
          char before[MAX_STR_LEN], *after = line+epos.Col;
          lstrcpy(before, line);
          before[epos.Col] = 0;
          if ( CheckMacroPos(lng, fm, before, after) )
            if ( checkMultipleChoice(lng, fm, before, after, 0) )
            {
              RunMacro(fm);
              redraw();
              return IGNORE_EVENT;
            }
        }
        return PROCESS_EVENT;
      }
      int fKey = FSF.FarInputRecordToKey(Rec);
      vState &= ( SHIFT_PRESSED | RIGHT_ALT_PRESSED | LEFT_ALT_PRESSED | RIGHT_CTRL_PRESSED | LEFT_CTRL_PRESSED );
      int spORret = ( fKey == KEY_ENTER || fKey == defExpandFKey );
      if ( !spORret && ( vState & ~SHIFT_PRESSED ) )
        return PROCESS_EVENT;
      int inImm = 0, inImmExp = 0;
      char vChar = Rec->Event.KeyEvent.uChar.AsciiChar;
      if ( !spORret )
      {
        if ( lng->ignoreCase )
        {
          char lwrStr[MAX_REG_LEN], c = (char)(FSF.LLower(vChar));
          inImm = lstrchr(StrLower(lstrcpy(lwrStr, lng->imm)), c) != NULL;
          inImmExp = lstrchr(StrLower(lstrcpy(lwrStr, lng->immExp)), c) != NULL;
        }
        else
        {
          inImm = lstrchr(lng->imm, vChar) != NULL;
          inImmExp = lstrchr(lng->immExp, vChar) != NULL;
        }
      }
      if ( ! ( spORret || ( vChar && ( inImm || inImmExp ) ) ) )
        return PROCESS_EVENT;
      if ( ( fKey == defExpandFKey ) || inImmExp )
      {
        EditorGetStr(&gs);
        return TryMacro(lng, gs, epos, line, inImmExp ? vChar : '\0') ? IGNORE_EVENT : PROCESS_EVENT;
      }
      else
      {
        int ret = PROCESS_EVENT;
        EditorGetStr(&gs);
        Editor2OEM(lstrcpy(line, gs.StringText));
        if ( !inComment(lng, line, epos.Col) )
        {
          if ( !spORret )
          {
            pluginBusy = 1;
            EditorProcessKey(vChar);
            pluginBusy = 0;
            ret = IGNORE_EVENT;
            EditorGetStr(&gs);
          }
          char *str = FirstNonSpace(gs.StringText);
          if ( str == NULL )
            return ret;
          int spacesCount = str-gs.StringText;
          FSF.Trim(lstrcpy(line, str));
          if ( lng->ignoreCase )
          {
            StrLower(line);
            vChar = (char)FSF.LLower(vChar);
          }
          int bounds[2];
          int j = FindIndent(spORret ? '\0' : vChar, lng, line, bounds);
          if ( j != -1 )
            return TryIndent(spORret, lng, j, line, str, spacesCount, bounds);
          return ret;
        }
      }
    }
  }
  return PROCESS_EVENT;
}

static inline char *delBackslash(char *s)
{
  if ( *s && lstrcmp(s, "\\") )
  {
    char *b = lstrchr(s, 0)-1;
    if ( *b == '\\' )
      *b = 0;
  }
  return s;
}

static char *QuoteText(char *Str)
{
  if ( *Str=='-' || *Str=='^' || lstrpbrk(Str," &+,") )
  {
    unsigned l = lstrlen(Str);
    memcpy(Str+1,Str,l);
    *Str = Str[l+1] = '\"';
    Str[l+2] = 0;
  }
  return Str;
}

static inline char *quote(int doQuote, char *Str)
{
  return doQuote ? QuoteText(Str) : Str;
}

static char* makeCmdLine(bool ask, char *buffer, const char *mask, const char *fn, const char *title = NULL, const char macro = '=')
{
  static char n1[NM], drv[NM], dir1[NM], fil1[NM], ext1[NM];
  static char n2[NM], tmp[NM], dir2[NM], fil2[NM], ext2[NM];
  char *i, *j, *e1 = ext1, *e2 = ext2, *smartQuote = NULL;
  fExpand(lstrcpy(tmp, fn));
  GetFullPathName(tmp, sizeof n1, n1, &j);
  fnSplit(n1, drv, dir1, fil1, ext1);
  GetShortPathName(tmp, n2, sizeof n2);
  fnSplit(n2, drv, dir2, fil2, ext2);
  lstrcat(lstrcat(lstrcat(lstrcpy(n1, drv), dir1), fil1), ext1);
  lstrcat(lstrcat(lstrcat(lstrcpy(n2, drv), dir2), fil2), ext2);
  if ( *e1 == '.' )
    e1++;
  if ( *e2 == '.' )
    e2++;
  i = (char*)mask;
  if ( ask )
  {
    nullUserStrings();
    if ( !scanUserInput(false, '=', mask, title) )
      return NULL;
  }
  *(j = buffer) = 0;
  while ( *i )
  {
    int dosName = 0, noDrive = 0, noBackslash = 0, doQuote = 1;
    char *m = lstrchr(i, macro);
    if ( m )
    {
      *m = 0;
      lstrcat(j, i);
      j = lstrchr(j, 0);
      *m = macro;
      if ( *++m == macro )
      {
        j[0] = macro;
        j[1] = 0;
        j = lstrchr(j, 0);
        i = m+1;
      }
      else if ( ask && lstrchr("0123456789?", *m) )
      {
        if ( *m != '?' )
          lstrcpy(j, userString[*m-'0']);
        j = lstrchr(j, 0);
        if ( m[1] == '\'' )
        {
          m++;
          while ( *++m && *m != '\'' )
            ;
        }
        i = m+1;
      }
      else
      {
        int modifyer = 1;
        while ( modifyer )
        {
          switch ( *m )
          {
            case 'd':
              dosName = 1;
              break;
            case 'm':
              noDrive = 1;
              break;
            case 't':
              noBackslash = 1;
              break;
            case 'o':
              doQuote = 0;
              break;
            default:
              m--;
              modifyer = 0;
              break;
          }
          m++;
        }
        switch ( *m )
        {
          case 'M':
            lstrcat(j, drv);
            break;
          case 'N':
            lstrcat(j, quote(doQuote, lstrcpy(tmp, dosName ? fil2 : fil1)));
            break;
          case 'E':
            lstrcat(j, quote(doQuote, lstrcpy(tmp, dosName ? e2 : e1)));
            break;
          case 'F':
            lstrcpy(tmp, dosName ? fil2 : fil1);
            if ( *(dosName ? e2 : e1) )
              lstrcat(lstrcat(tmp, "."), dosName ? e2 : e1);
            lstrcat(j, quote(doQuote, tmp));
            break;
          case 'D':
            if ( noDrive )
              *tmp = 0;
            else
              lstrcpy(tmp, drv);
            lstrcat(tmp, dosName ? dir2 : dir1);
            if ( noBackslash )
              delBackslash(tmp);
            lstrcat(j, quote(doQuote, tmp));
            break;
          case 'P':
            if ( noDrive )
            {
              lstrcpy(tmp, dosName ? dir2 : dir1);
              lstrcat(tmp, dosName ? fil2 : fil1);
              if ( *(dosName ? e2 : e1) )
                lstrcat(lstrcat(tmp, "."), dosName ? e2 : e1);
              lstrcat(j, quote(doQuote, tmp));
            }
            else
              lstrcat(j, quote(doQuote, lstrcpy(tmp, dosName ? n2 : n1)));
            break;
          case '\'':
            if ( smartQuote )
            {
              for ( char *p = smartQuote ; *p ; p++ )
                if ( *p == '\"' )
                  *p = '\'';
              quote(1, smartQuote);
              smartQuote = NULL;
            }
            else
              smartQuote = lstrchr(j, 0);
            break;
        }
        i = m+1;
        j = lstrchr(j, 0);
      }
    }
    else
      break;
  }
  lstrcat(j, i);
  return buffer;
}

static FarMenuItemEx *compilerOut = NULL;
static TCollection *errColl = NULL;
static int compilerOutN = 0;
static int compilerOutP = -1;

void WINAPI _export ExitFAR()
{
  DoneMacro();
  if ( eList )
  {
    eList->done();
    delete eList;
  }
  if ( compilerOut )
  {
    delete [] compilerOut;
    compilerOut = NULL;
  }
  if ( errColl )
  {
    errColl->done();
    delete errColl;
  }
}

struct TErrData
{
  char fn[NM];
  int line, col, msgCount;
  char message[64][64];
  char colSearch[MAX_REG_LEN];
};

static char cmd[MAX_STR_LEN];

static bool showErrorMsgBox(int res)
{
  int n = (int)compilerOut[res].UserData-1;
  if ( errColl && ( n >= 0 ) && ( n < (int)errColl->getCount() ) )
  {
    TErrData *err = (TErrData*)((*errColl)[n]);
    const char *MsgItems[] =
    {
      cmd, err->fn, "\x01",
      err->message[ 0], err->message[ 1], err->message[ 2], err->message[ 3], err->message[ 4],
      err->message[ 5], err->message[ 6], err->message[ 7], err->message[ 8], err->message[ 9],
      err->message[10], err->message[11], err->message[12], err->message[13], err->message[14],
      err->message[15], err->message[16], err->message[17], err->message[18], err->message[19]
    };
    redraw();
    Info.Message(Info.ModuleNumber, FMSG_MB_OK|FMSG_LEFTALIGN, NULL, MsgItems, err->msgCount+3, 0);
    return true;
  }
  return false;
}

static void jumpToError(int aj, bool showMsgBox)
{
  if ( compilerOut && ( aj >= 0 ) && ( aj <= compilerOutN ) )
  {
    int n = (int)compilerOut[aj].UserData-1;
    if ( errColl && ( n >= 0 ) && ( n < (int)errColl->getCount() ) )
    {
      TErrData *err = (TErrData*)((*errColl)[n]);
      int nLine = -1, nCol = err->col, sLine = -1;
      if ( err->line > 0 )
        nLine = err->line-1;
      if ( err->col == 0 )
        nCol--;
      char errfn[NM];
      fExpand(lstrcpy(errfn, err->fn ? err->fn : ""));
      WindowInfo wi;
      bool found = false;
      int n = Info.AdvControl(Info.ModuleNumber, ACTL_GETWINDOWCOUNT, NULL);
      for ( int i = -1 ; i < n ; i++ )
      {
        wi.Pos = i;
        Info.AdvControl(Info.ModuleNumber, ACTL_GETWINDOWINFO, (void*)&wi);
        if ( wi.Type == WTYPE_EDITOR )
        {
          if ( !FSF.LStricmp(errfn, wi.Name) )
          {
            if ( i != -1 )
            {
              Info.AdvControl(Info.ModuleNumber, ACTL_SETCURRENTWINDOW, (void*)i);
              Info.AdvControl(Info.ModuleNumber, ACTL_COMMIT, NULL);
            }
            sLine = nLine;
            found = true;
            break;
          }
        }
      }
      if ( !found )
        Info.Editor(err->fn, NULL, 0, 0, -1, -1, EF_NONMODAL|EF_ENABLE_F6|EF_IMMEDIATERETURN, err->line, nCol);
      EditorSetPosEx(sLine, nCol, err->colSearch);
      if ( showMsgBox )
        showErrorMsgBox(aj);
    }
  }
}

static void showCompileOut()
{
  if ( compilerOut )
  {
    int Code, BreakKeys[] = { VK_RETURN, VK_F4, VK_F3, 0 };
    if ( compilerOutP >= 0 && compilerOutP < compilerOutN )
    {
      for ( int i = 0 ; i < compilerOutN ; i++ )
        compilerOut[i].Flags &= ~MIF_SELECTED;
      compilerOut[compilerOutP].Flags |= MIF_SELECTED;
    }
    for ( ; ; )
    {
      compilerOutP = Info.Menu(Info.ModuleNumber, -1, -1, 0,
        FMENU_WRAPMODE|FMENU_SHOWAMPERSAND|FMENU_USEEXT,
        cmd, GetMsg(MMenuBottom), NULL, BreakKeys, &Code,
        (const FarMenuItem *)compilerOut, compilerOutN);
      if ( compilerOutP < 0 )
        break;
      if ( Code == 2 )
      {
        if ( compilerOutP <= compilerOutN && showErrorMsgBox(compilerOutP) )
        {
          for ( int i = 0 ; i < compilerOutN ; i++ )
            compilerOut[i].Flags &= ~MIF_SELECTED;
          compilerOut[compilerOutP].Flags |= MIF_SELECTED;
        }
      }
      else
      {
        jumpToError(compilerOutP, false);
        break;
      }
    }
  }
}

static void execute(TCollection *coll, const char* title, bool showExec)
{
  HANDLE hScreen = Info.SaveScreen(0,0,-1,-1);
  if ( !showExec )
  {
    const char *MsgItems[]={GetMsg(MTitle), title};
    Info.Message(Info.ModuleNumber,0,NULL,MsgItems,dimOf(MsgItems),0);
  }
  ExecConsoleApp(cmd, coll, true, showExec);
  if ( showExec )
    Info.Control(NULL, FCTL_SETUSERSCREEN, NULL);
  Info.RestoreScreen(hScreen);
  redraw();
}

static bool SaveAll()
{
  bool res = true;
  WindowInfo wi;
  wi.Pos = -1;
  Info.AdvControl(Info.ModuleNumber, ACTL_GETWINDOWINFO, (void*)&wi);
  int home = wi.Pos;
  int n = Info.AdvControl(Info.ModuleNumber, ACTL_GETWINDOWCOUNT, NULL);
  for ( int i = 0 ; i < n ; i++ )
  {
    wi.Pos = i;
    Info.AdvControl(Info.ModuleNumber, ACTL_GETWINDOWINFO, (void*)&wi);
    if ( wi.Type == WTYPE_EDITOR && wi.Modified )
    {
      if ( Info.AdvControl(Info.ModuleNumber, ACTL_SETCURRENTWINDOW, (void*)i) )
      {
        Info.AdvControl(Info.ModuleNumber, ACTL_COMMIT, NULL);
        if ( !Info.EditorControl(ECTL_SAVEFILE, NULL) )
          res = false;
      }
      else
        res = false;
    }
  }
  Info.AdvControl(Info.ModuleNumber, ACTL_SETCURRENTWINDOW, (void*)home);
  Info.AdvControl(Info.ModuleNumber, ACTL_COMMIT, NULL);
  return res;
}

static bool findBaseFile(const char *file, char *baseFile)
{
  bool found = false;
  char testFile[NM];
  int n = Info.AdvControl(Info.ModuleNumber, ACTL_GETWINDOWCOUNT, NULL);
  GetCurrentDirectory(sizeof(testFile), testFile);
  FSF.AddEndSlash(testFile);
  lstrcat(testFile, FSF.PointToName(file));
  if ( isFile(testFile) )
  {
    if ( baseFile )
      lstrcpy(baseFile, testFile);
    found = true;
  }
  if ( !found )
  {
    for ( int i = -1 ; i < n ; i++ )
    {
      WindowInfo wi;
      wi.Pos = i;
      Info.AdvControl(Info.ModuleNumber, ACTL_GETWINDOWINFO, (void*)&wi);
      if ( wi.Type == WTYPE_EDITOR )
      {
        char *p = (char*)lstrrchr(lstrcpy(testFile, wi.Name), '\\');
        if ( p )
        {
          lstrcpy(p+1, FSF.PointToName(file));
          if ( isFile(testFile) )
          {
            if ( baseFile )
              lstrcpy(baseFile, testFile);
            found = true;
            break;
          }
        }
      }
    }
  }
  return found;
}

static bool validMenuItem(const char *fn, TExec *e)
{
  char baseFile[NM] = "";
  bool enable = true;
  if ( e->enable[0] )
  {
    makeCmdLine(false, baseFile, e->enable, fn);
    enable = isFile(baseFile);
    if ( !enable && e->searchBase )
      enable = findBaseFile(baseFile, NULL);
  }
  if ( enable && e->disable[0] )
    enable = !isFile(makeCmdLine(false, baseFile, e->disable, fn));
  return enable;
}

static char *makeTitle(const char *line, size_t size, const char *fn)
{
  static char temp[MAX_STR_LEN];
  makeCmdLine(false, temp, line, fn);
  FSF.TruncStr(temp, size);
  return temp;
}

static bool parseError(TLang *lng, const char *compiler, const char *fn, char *line, unsigned ci, int& aj)
{
  int lineBounds[2], colBounds[2], fileBounds[2];
  TCompiler *e = NULL;
  bool found = false, res = false;
  for ( unsigned i = 0 ; i < lng->compilerColl.getCount() ; i++ )
  {
    e = (TCompiler*)(lng->compilerColl[i]);
    if ( e && e->err[0] && !FSF.LStricmp(compiler, e->title) )
    {
      int bounds[3][2] = { { 0, 0 }, { 0, 0 }, { 0, 0 } };
      int pos[3] = { e->line, e->col, e->fileMatch };
      if ( strMatch(line, e->err, "/", ".*/i", 3, bounds, pos) )
      {
        for ( int j = 0 ; j < 2 ; j++ )
        {
          lineBounds[j] = bounds[0][j];
          colBounds[j]  = bounds[1][j];
          fileBounds[j] = bounds[2][j];
        }
        found = true;
        break;
      }
    }
  }
  if ( found && e && e->err[0] )
  {
    int len, lineNo = -1, colNo = -1;
    compilerOut[ci].Flags = 0;
    char fileName[NM], colSearch[MAX_REG_LEN] = "";
    if ( e->line >= 0 )
    {
      len = lineBounds[1]-lineBounds[0];
      lstrcpyn(fileName, line+lineBounds[0], len+1);
      fileName[len] = 0;
      if ( len )
        lineNo = FSF.atoi(fileName);
    }
    if ( e->col >= 0 )
    {
      len = colBounds[1]-colBounds[0];
      lstrcpyn(fileName, line+colBounds[0], len+1);
      fileName[len] = 0;
      if ( len )
      {
        if ( e->searchCol )
        {
          colNo = -2;
          lstrcpy(colSearch, fileName);
        }
        else
          colNo = FSF.atoi(fileName);
      }
    }
    if ( e->fileMatch >= 0 )
    {
      len = fileBounds[1]-fileBounds[0];
      lstrcpyn(fileName, line+fileBounds[0], len+1);
      fileName[len] = 0;
      if ( !len )
        lstrcpy(fileName, fn);
    }
    else
      lstrcpy(fileName, fn);
    TErrData *errData = new TErrData;
    fExpand(lstrcpy(errData->fn, fileName));
    char *p = line;
    errData->msgCount = 0;
    for ( ; ; )
    {
      char brk[260];
      Info.AdvControl(Info.ModuleNumber, ACTL_GETSYSWORDDIV, brk);
      lstrcpyn(errData->message[errData->msgCount], p, 64);
      errData->message[errData->msgCount][63] = 0;
      if ( lstrpbrk(errData->message[errData->msgCount], brk) )
      {
        char *pp = lstrchr(errData->message[errData->msgCount], 0);
        while ( !lstrchr(brk, *(--pp)) )
          ;
        pp[1] = 0;
      }
      p += lstrlen(errData->message[errData->msgCount]);
      FSF.Trim(errData->message[errData->msgCount]);
      if ( !errData->message[errData->msgCount][0] )
      {
        lstrcpy(errData->message[errData->msgCount], "\x01");
        errData->msgCount++;
        break;
      }
      errData->msgCount++;
    }
    errData->line = lineNo;
    errData->col = colNo;
    lstrcpy(errData->colSearch, colSearch);
    errColl->insert(errData);
    compilerOut[ci].UserData = errColl->getCount();
    if ( aj < 0 )
    {
      compilerOut[ci].Flags = MIF_SELECTED;
      aj = (int)ci;
    }
    res = true;
  }
  return res;
}

static bool runCompiler(TLang *lng, const char *fn, const char *title, TExec *e)
{
  char cwd[NM] = "",  baseFile[NM] = "";
  if ( e->enable[0] )
  {
    char temp[NM];
    makeCmdLine(false, baseFile, e->enable, fn);
    bool enable = isFile(baseFile);
    if ( !enable && e->searchBase && findBaseFile(baseFile, temp) )
      lstrcpy(baseFile, temp);
  }
  if ( !makeCmdLine(true, cmd, e->cmd, fn, makeTitle(title, 30, fn)) )
    return false;
  FSF.ExpandEnvironmentStr(cmd, cmd, sizeof(cmd));
  if ( e->saveType != saNone )
  {
    if ( e->saveType == saCurrent )
    {
      WindowInfo wi;
      wi.Pos = -1;
      Info.AdvControl(Info.ModuleNumber, ACTL_GETWINDOWINFO, (void*)&wi);
      if ( wi.Type == WTYPE_EDITOR && wi.Modified )
        if ( !Info.EditorControl(ECTL_SAVEFILE, NULL) )
          return false;
    }
    else
    {
      if ( !SaveAll() )
      {
        const char *MsgItems[]={GetMsg(MTitle), GetMsg(MSaveError), GetMsg(MContinue)};
        if ( Info.Message(Info.ModuleNumber, FMSG_MB_YESNO, NULL, MsgItems, dimOf(MsgItems), 0) )
          return false;
      }
    }
  }
  if ( e->cd != cdNone )
  {
    char newDir[NM] = "", *p;
    if ( e->cd == cdFile )
      p = (char*)lstrrchr(lstrcpy(newDir, fn), '\\');
    else if ( ( e->cd == cdBase ) && *baseFile )
      p = (char*)lstrrchr(lstrcpy(newDir, baseFile), '\\');
    if ( p )
      *p = 0;
    if ( *newDir )
    {
      GetCurrentDirectory(sizeof(cwd), cwd);
      SetCurrentDirectory(newDir);
    }
  }
  if ( errColl )
    errColl->removeAll();
  else
    errColl = new TCollection;
  TCollection compilerColl;
  compilerOutP = -1;
  execute(&compilerColl, cmd, e->echo);
  unsigned n = compilerColl.getCount();
  if ( n )
  {
    int nErr = 0, aj = -1;
    compilerOut = new FarMenuItemEx[compilerOutN = n+1];
    for ( unsigned i = 0 ; i < n ; i++ )
    {
      char *line = (char*)(compilerColl[i]);
      lstrcpyn(compilerOut[i].Text.Text, line, sizeof(compilerOut[i].Text.Text));
      compilerOut[i].Text.Text[sizeof(compilerOut[i].Text.Text)-1] = 0;
      compilerOut[i].Flags = MIF_DISABLE;
      compilerOut[i].UserData = 0;
      char compiler[MAX_STR_LEN];
      lstrcpy(compiler, e->compiler);
      char *pp, *p = compiler;
      bool found = false;
      while ( ( pp = lstrchr(p, ',') ) != NULL )
      {
        *pp = 0;
        if ( parseError(lng, p, fn, line, i, aj) )
        {
          nErr++;
          found = true;
        }
        p = pp+1;
        if ( found )
          break;
      }
      if ( !found && parseError(lng, p, fn, line, i, aj) )
        nErr++;
    }
    if ( ( aj >= 0 ) && ( ( e->jumpType == jtFirst ) ) || ( ( e->jumpType == jtSmart ) && ( nErr == 1 ) ) )
      jumpToError(aj, true);
    else
    {
      if ( aj < 0 )
      {
        if ( compilerOut[n-1].Text.Text[0] == 0 )
        {
          compilerOutN--;
          n--;
        }
        compilerOut[n].Text.Text[0] = 0;
        compilerOut[n].Flags = MIF_SELECTED;
      }
      else
        compilerOutN--;
      if ( ( e->jumpType == jtMenu ) || ( e->jumpType == jtSmart ) )
        showCompileOut();
    }
  }
  else if ( compilerOut )
  {
    delete [] compilerOut;
    compilerOut = NULL;
  }
  compilerColl.done();
  if ( *cwd )
    SetCurrentDirectory(cwd);
  return true;
}

static void SelectTemplate(TEInfo *te)
{
  unsigned lc = langColl.getCount();
  FarMenuItem *amenu = new FarMenuItem[lc+2];
  if ( amenu )
  {
    for ( unsigned i = 0 ; i < lc ; i++ )
    {
      FSF.sprintf(amenu[i].Text, "%s", ((TLang*)langColl[i])->desc);
      amenu[i].Selected = (unsigned)(te->lang) == i;
      amenu[i].Checked = amenu[i].Separator = false;
    }
    amenu[lc].Selected = amenu[lc].Checked = false;
    amenu[lc].Separator = true;
    *amenu[lc].Text = 0;
    FSF.sprintf(amenu[lc+1].Text, "%s", GetMsg(MOtherLang));
    amenu[lc+1].Selected = te->lang == -1;
    amenu[lc+1].Checked = amenu[lc+1].Separator = false;
    int res = Info.Menu(Info.ModuleNumber,-1,-1,0,FMENU_WRAPMODE|FMENU_AUTOHIGHLIGHT,GetMsg(MSelectLang),NULL,NULL,NULL,NULL,amenu,lc+2);
    if ( res != -1 )
    {
      if ( pluginStop )
        SetRegKey(HKEY_CURRENT_USER, "", "Disable", pluginStop = 0);
      te->lang = ( (unsigned)res < lc ) ? res : -1;
    }
    delete [] amenu;
  }
}

static bool sep = true;

static void addToCompilerMenu(const char *line, FarMenuItemEx *amenu, int& i, DWORD j, const char *fn, DWORD AccelKey)
{
  FSF.sprintf(amenu[i].Text.Text, "%s", makeTitle(line, sizeof(amenu[i].Text.Text)-1, fn));
  amenu[i].UserData = j;
  amenu[i].Flags = 0;
  amenu[i].AccelKey = AccelKey;
  if ( amenu[i].Text.Text[0] )
  {
    i++;
    sep = false;
  }
  else
  {
    if ( !sep )
    {
      amenu[i].Flags = MIF_SEPARATOR;
      i++;
    }
    sep = true;
  }
}

static void CompilerMenu(const char *Name, int lang)
{
  char top[NM], FileName[NM];
  lstrcpy(FileName, Name);
  lstrcpy(top, GetMsg(MTitle));
  TLang *lng = (TLang*)(langColl[lang]);
  unsigned ec = 0;
  if ( lng )
    ec = lng->execColl.getCount();
  FarMenuItemEx *amenu = new FarMenuItemEx[ec+2];
  if ( amenu )
  {
    const DWORD scId = -1;
    int i = 0;
    sep = true;
    if ( ec )
    {
      for ( unsigned j = 0 ; j < ec ; j++ )
      {
        TExec *ex = ((TExec*)lng->execColl[j]);
        if ( validMenuItem(FileName, ex) )
          addToCompilerMenu(ex->title, amenu, i, j, FileName, 0);
      }
      addToCompilerMenu("", amenu, i, 0, FileName, 0);
    }
    addToCompilerMenu(GetMsg(MShowOutput), amenu, i, scId, FileName, KEY_MULTIPLY);
    if ( !compilerOut )
      amenu[i-1].Flags = MIF_DISABLE;
    for ( int k = 0 ; k < i ; k++ )
      if ( ! ( amenu[k].Flags & ( MIF_DISABLE|MIF_SEPARATOR ) ) )
      {
        amenu[k].Flags |= MIF_SELECTED;
        break;
      }
    int res = Info.Menu(Info.ModuleNumber,-1,-1,0,FMENU_WRAPMODE|FMENU_AUTOHIGHLIGHT|FMENU_USEEXT,top,NULL,NULL,NULL,NULL,(const FarMenuItem *)amenu, i);
    if ( res != -1 )
    {
      DWORD id = amenu[res].UserData;
      if ( id == scId )
        showCompileOut();
      else if ( lng )
      {
        TExec *e = (TExec*)lng->execColl[id];
        runCompiler(lng, FileName, e->title, e);
      }
    }
    delete [] amenu;
  }
}

static void SelectCompiler(TEInfo *te)
{
  EditorInfo ei;
  Info.EditorControl(ECTL_GETINFO, &ei);
  CompilerMenu(ei.FileName, te->lang);
}

HANDLE WINAPI _export OpenPlugin(int OpenFrom, int item)
{
  if ( !IsOldFar )
  {
    int n;
    EditorInfo ei;
    WindowInfo wi;
    switch ( OpenFrom )
    {
      case OPEN_VIEWER:
        wi.Pos = -1;
        Info.AdvControl(Info.ModuleNumber, ACTL_GETWINDOWINFO, &wi);
        n = findLngID(wi.Name);
        if ( n != -1 )
//          if ( ((TLang*)(langColl[n]))->panel )
          CompilerMenu(wi.Name, n);
        break;
      case OPEN_SHORTCUT:
      case OPEN_COMMANDLINE:
        n = findLngID((char*)item);
        if ( n != -1 )
//          if ( ((TLang*)(langColl[n]))->panel )
          CompilerMenu((char*)item, n);
        break;
      case OPEN_EDITOR:
        Info.EditorControl(ECTL_GETINFO, &ei);
        initEList();
        n = eList->findID(ei.EditorID);
        if ( n == -1 )
          n = eListInsert(&ei);
        if ( n != -1 )
        {
          TEInfo *te = (TEInfo*)(*eList)[n];
          if ( te )
          {
            FarMenuItem mMenu[2];
            int count = dimOf(mMenu);
            for ( int i = 0 ; i < count ; i++ )
              mMenu[i].Selected = mMenu[i].Checked = mMenu[i].Separator = false;
            lstrcpy(mMenu[0].Text, GetMsg(MMenuSelectLang));
            lstrcpy(mMenu[1].Text, GetMsg(MMenuCompile));
            int mode = Info.Menu(Info.ModuleNumber,-1,-1,0,FMENU_WRAPMODE,GetMsg(MTitle),NULL,NULL,NULL,NULL,mMenu,count);
            if ( mode == 0 )
              SelectTemplate(te);
            else if ( mode == 1 )
              SelectCompiler(te);
          }
        }
        break;
    }
  }
  return INVALID_HANDLE_VALUE;
}
